package com.mycomp.search.servlet;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.StringTokenizer;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.hadoop.conf.Configuration;
import org.apache.nutch.searcher.Hit;
import org.apache.nutch.searcher.HitDetails;
import org.apache.nutch.searcher.Hits;
import org.apache.nutch.searcher.NutchBean;
import org.apache.nutch.searcher.Query;
import org.apache.nutch.searcher.Summary;
import org.apache.nutch.util.NutchConfiguration;
import org.json.JSONArray;
import org.json.JSONObject;
import org.json.XML;

/**
 * A servlet that searches the index and returns the search results in text, xml
 * or json format
 * 
 * json is the default format
 * 
 * @author ravi.chintakunta
 * 
 */
public class SearchServlet extends HttpServlet {

	private static final long serialVersionUID = -7166599319150821314L;

	private Configuration nutchConf;

	List<String> indexKeysList;

	public void init(ServletConfig config) throws ServletException {
		super.init(config);
		indexKeysList = new ArrayList<String>();
		nutchConf = NutchConfiguration.get(config.getServletContext());
		Enumeration enumeration = config.getServletContext()
				.getInitParameterNames();
		while (enumeration.hasMoreElements()) {
			indexKeysList.add(enumeration.nextElement().toString());
		}
	}

	private boolean isValidIndexKey(String[] indexKeys) {
		for (String indexKey : indexKeys) {
			if (!indexKeysList.contains(indexKey)) {
				return false;
			}
		}
		return true;
	}

	private String joinIndexKeys(String[] indexKeys) {
		if (indexKeys.length == 1) {
			return indexKeys[0];
		}
		StringBuilder builder = new StringBuilder();
		Arrays.sort(indexKeys);
		for (String key : indexKeys) {
			builder.append(key);
			builder.append(":");
		}
		return builder.toString();
	}

	protected void doGet(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		String[] indexKeys = request.getParameterValues("index");
		String indexKey;
		NutchBean bean;
		if (indexKeys == null || !isValidIndexKey(indexKeys)) {
			indexKey = null;
			bean = NutchBean.get(getServletContext(), nutchConf);
		} else {
			indexKey = joinIndexKeys(indexKeys);
			bean = NutchBean.get(getServletContext(), nutchConf, indexKey);
		}
		// set the character encoding to use when interpreting request values
		request.setCharacterEncoding("UTF-8");

		String queryString;
		if (request.getParameter("advanced") != null) {
			String all = request.getParameter("all").trim();
			String exact = request.getParameter("exact").trim();
			if (exact.length() > 0) {
				exact = "\"" + exact + "\"";
			}
			String atleast = request.getParameter("atleast").trim();
			if (atleast.trim().length() > 0) {
				StringTokenizer tokenizer = new StringTokenizer(atleast);
				StringBuffer buffer = new StringBuffer();
				while (tokenizer.hasMoreElements()) {
					if (buffer.length() > 0) {
						buffer.append(" OR ");
					}
					buffer.append(tokenizer.nextElement());
				}
				atleast = buffer.toString();
			}
			String without = request.getParameter("without").trim();
			if (without.trim().length() > 0) {
				StringTokenizer tokenizer = new StringTokenizer(without);
				StringBuffer buffer = new StringBuffer();
				while (tokenizer.hasMoreElements()) {
					buffer.append("-");
					buffer.append(tokenizer.nextElement());
					buffer.append(" ");
				}
				without = buffer.toString();
			}
			queryString = all + " " + exact + " " + atleast + " " + without;
		} else {
			// get query from request
			queryString = request.getParameter("query");
			if (queryString == null)
				queryString = "";
		}
		String format = "json";
		if (request.getParameter("format") != null) {
			format = request.getParameter("format");
		}

		int start = 0; // first hit to display
		String startString = request.getParameter("start");
		if (startString != null)
			start = Integer.parseInt(startString);

		int hitsPerPage = 10; // number of hits to display
		String hitsString = request.getParameter("hitsPerPage");
		if (hitsString != null)
			hitsPerPage = Integer.parseInt(hitsString);

		int hitsPerSite = 0; // max hits per site
		String hitsPerSiteString = request.getParameter("hitsPerSite");
		if (hitsPerSiteString != null)
			hitsPerSite = Integer.parseInt(hitsPerSiteString);

		String sort = request.getParameter("sort");
		boolean reverse = sort != null
				&& "true".equals(request.getParameter("reverse"));

		Query query = Query.parse(queryString, "", nutchConf);

		int hitsToRetrieve = hitsPerPage;

		Hits hits;
		long startTime = System.currentTimeMillis();
		String suggestion = "";
		try {
			if (indexKeys == null) {
				hits = bean.search(query, start + hitsToRetrieve, hitsPerSite,
						"site", sort, reverse);
			} else {
				hits = bean.search(query, start + hitsToRetrieve, hitsPerSite,
						"site", sort, reverse, indexKey);
			}
			if (start == 0 && hits.getLength() < 5
					&& queryString.split(" ").length == 1) {
				String[] suggestions = bean.suggest(queryString);
				if (suggestions != null && suggestions.length > 0) {
					suggestion = suggestions[0];

				}
			}
		} catch (IOException e) {
			hits = new Hits(0, new Hit[0]);
		}
		int end = (int) Math.min(hits.getLength(), start + hitsPerPage);
		int length = end - start;
		int realEnd = (int) Math.min(hits.getLength(), start + hitsToRetrieve);

		Hit[] show = hits.getHits(start, realEnd - start);
		HitDetails[] details;
		Summary[] summaries;
		if (indexKey == null) {
			details = bean.getDetails(show);
			summaries = bean.getSummary(details, query);
		} else {
			details = bean.getDetails(show, indexKey);
			summaries = bean.getSummary(details, query, indexKey);

		}

		int time = (int) (System.currentTimeMillis() - startTime);

		String ip = request.getParameter("remoteip") == null ? request
				.getRemoteAddr() : request.getParameter("remoteip");
		String user = request.getParameter("user") == null ? "" : request
				.getParameter("user");
		String team = request.getParameter("team") == null ? "" : request
				.getParameter("team");
		StringBuffer logBuffer = new StringBuffer();
		logBuffer.append(">>> ip:");
		logBuffer.append(ip);
		logBuffer.append("; user:");
		logBuffer.append(user);
		logBuffer.append("; team:");
		logBuffer.append(team);
		logBuffer.append("; query:");
		logBuffer.append(queryString);
		logBuffer.append("; hits:");
		logBuffer.append(hits.getTotal());
		NutchBean.LOG.info(logBuffer.toString());
		response.setContentType("text/plain");
		PrintWriter writer = response.getWriter();
		JSONObject root = null;
		JSONArray results = null;
		if (!format.equals("text")) {
			root = new JSONObject();
			root.put("start", start);
			root.put("end", end);
			root.put("query", queryString);
			root.put("total", hits.getTotal());
			root.put("time", time);
			if (suggestion.trim().length() > 0) {
				root.put("suggestion", suggestion);
			}
			if (queryString.trim().length() == 0) {
				root.put("status_code", SearchConstants.NO_QUERY_CODE);
				root.put("status_message", SearchConstants.NO_QUERY_MESSAGE);
			} else if (hits.getTotal() == 0) {
				root.put("status_code", SearchConstants.NO_RESULTS_CODE);
				root.put("status_message", SearchConstants.NO_RESULTS_MESSAGE);
			} else {
				root.put("status_code", SearchConstants.OK_CODE);
				root.put("status_message", SearchConstants.OK_MESSAGE);
			}
			results = new JSONArray();
		} else {
			writer.write("Hits " + start + " - " + end + " of about "
					+ hits.getTotal() + " matching pages  (" + time
					+ " milliseconds)\n\n");
			if (suggestion.trim().length() > 0) {
				writer.write("Did you mean? " + suggestion + " \n\n");
			}
		}
		for (int i = 0; i < length; i++) { // display the hits
			HitDetails detail = details[i];
			String title = detail.getValue("title");
			String url = detail.getValue("url");
			String summary = summaries[i].toHtml(true);
			if (title == null || title.equals("")) {
				title = url;
			}
			if (format.equals("text")) {
				writer.write(title + "\n");
				writer.write(summary + "\n");
				writer.write(url + "\n");
				writer.write("\n");
			} else {
				JSONObject result = new JSONObject();
				result.put("title", title);
				result.put("summary", summary);
				result.put("url", url);
				results.put(result);
			}
		}
		/* get the hit count for each index key */
		JSONArray counts = new JSONArray();
		for (String key : indexKeys) {
			bean = NutchBean.get(getServletContext(), nutchConf, key);
			hits = bean.search(query, 10, 0, "site", sort, reverse, key);
			if (format.equals("text")) {
				writer.write(key);
				writer.write(" : ");
				writer.write(String.valueOf(hits.getTotal()));
				writer.write("\n");
			} else {
				JSONObject count = new JSONObject();
				count.put(key, hits.getTotal());
				counts.put(count);
			}
		}

		if (!format.equals("text")) {
			root.put("counts", counts);
			root.put("results", results);
		}
		if (format.equals("json")) {
			writer.write(root.toString());
		}
		if (format.equals("xml")) {
			writer.write(XML.toString(root, "root"));
		}
		writer.flush();

	}

	protected void doPost(HttpServletRequest arg0, HttpServletResponse arg1)
			throws ServletException, IOException {
		doGet(arg0, arg1);
	}

	@Override
	public void destroy() {
		super.destroy();
		try {
			NutchBean bean = NutchBean.get(getServletContext(), nutchConf,
					indexKeysList.get(0).toString());
			bean.close();
		} catch (Exception e) {

		}
	}

}
